from raichu.class_domain import make_scaffold_domain
from pikachu.general import read_smiles
from pikachu.reactions.functional_groups import (
    BondDefiner,
    GroupDefiner,
    combine_structures,
)
from raichu.reactions.general import initialise_atom_attributes


_STARTER_PKS_TO_SULPHUR = {
    "WILDCARD_STARTER": 0,
    "PROPIONYL_COA": 0,
    "ACETYL_COA": 0,
    "BENZOYL_COA": 0,
    "METHYL_BUTYRYL_COA_3": 0,
    "METHYL_BUTYRYL_COA_2": 0,
    "TRANS_CYCLOPENTANE_DICARBOXYL_COA": 0,
    "CYCLOHEXANE_CARBOXYL_COA": 0,
    "HYDROXY_MALONYL_COA_2": 0,
    "HYDROXY_MALONYL_COA_2R": 0,
    "HYDROXY_MALONYL_COA_2S": 0,
    "CHLOROETHYL_MALONYL_COA": 0,
    "ISOBUTYRYL_COA": 0,
    "GLYCINE": 0,
    "HYDROXY_PROPENOYL_COA_3_23E": 5,
    "HYDROXY_BUTENOYL_COA_3_23E": 5,
    "DIHYDROXY_BUTANOLYL_COA_2R3": 8,
    "TRIHYDROXY_PROPANOLYL_COA_233": 7,
    "O_METHYLACETYL_COA": 3,
    "HYDROXY_PROPENOYL_COA_3_23Z": 5,
    "OXOMALONYL_COA_2": 6,
    "METHYL_HYDROXY_PROPENOYL_COA_2_3_23Z": 6,
    "DIHYDROXY_BUTANOLYL_COA_23": 7,
    "DIHYDROXY_BUTANOLYL_COA_2S3S": 9,
    "HEPTATRIENOYL_COA": 0,
    "HYDROXYPROPIONYL_COA_2R": 5,
    "DIHYDROXY_PROPANOLYL_COA_33": 6,
    "LACTYL_COA": 5,
    "PHENYLACETYLCOA": 0,
    "METHOXYFORMYL_COA": 3,
}


_STARTER_PKS_TO_BACKBONE = {
    "WILDCARD_STARTER": [1, 2],
    "PROPIONYL_COA": [1, 3, 4],
    "ACETYL_COA": [1, 2],
    "BENZOYL_COA": [1, 2, 3],
    "METHYL_BUTYRYL_COA_3": [1, 3, 4, 5],
    "METHYL_BUTYRYL_COA_2": [1, 3, 5, 6],
    "TRANS_CYCLOPENTANE_DICARBOXYL_COA": [1, 2, 4, 9, 11],
    "CYCLOHEXANE_CARBOXYL_COA": [1, 3, 4, 5],
    "HYDROXY_MALONYL_COA_2": [1, 3, 4, 6],
    "HYDROXY_MALONYL_COA_2R": [1, 3, 5, 7],
    "HYDROXY_MALONYL_COA_2S": [1, 3, 5, 7],
    "CHLOROETHYL_MALONYL_COA": [1, 2, 6, 7, 8],
    "ISOBUTYRYL_COA": [1, 3, 4],
    "GLYCINE": [1, 2, 3],
    "HYDROXY_PROPENOYL_COA_3_23E": [1, 2, 3, 4],
    "HYDROXY_BUTENOYL_COA_3_23E": [1, 2, 3, 4],
    "DIHYDROXY_BUTANOLYL_COA_2R3": [0, 1, 3, 6],
    "TRIHYDROXY_PROPANOLYL_COA_233": [0, 1, 3, 5],
    "O_METHYLACETYL_COA": [0, 1, 2],
    "HYDROXY_PROPENOYL_COA_3_23Z": [1, 3, 4],
    "OXOMALONYL_COA_2": [0, 1, 3, 5],
    "METHYL_HYDROXY_PROPENOYL_COA_2_3_23Z": [1, 3, 5],
    "DIHYDROXY_BUTANOLYL_COA_23": [0, 1, 3, 5],
    "DIHYDROXY_BUTANOLYL_COA_2S3S": [0, 1, 4, 7],
    "HEPTATRIENOYL_COA": [1, 3, 4, 5, 6, 7, 8],
    "HYDROXYPROPIONYL_COA_2R": [0, 1, 4],
    "DIHYDROXY_PROPANOLYL_COA_33": [0, 1, 3, 4],
    "LACTYL_COA": [0, 1, 4],
    "PHENYLACETYLCOA": [1, 3, 4, 5, 6],
    "METHOXYFORMYL_COA": [0, 1, 2],
}


class PksStarterUnit:
    def __init__(self, name, smiles, central_chain_indices):
        self.name = name
        self.smiles = smiles
        self.structure = read_smiles(self.smiles)
        initialise_atom_attributes(self.structure)
        self.structure.refresh_structure()
        self.sulphur = self.structure.atoms[_STARTER_PKS_TO_SULPHUR[self.name]]

        for atom_index in central_chain_indices:
            self.structure.atoms[atom_index].annotations.in_central_chain = True

    def attach_to_acp(self):
        domain = make_scaffold_domain("ACP")
        sh_bond = domain.bond_lookup[domain.atoms[1]][domain.atoms[2]]
        hydrogen = domain.atoms[2]
        sulphur_1 = domain.atoms[1]
        carbon = self.sulphur.get_neighbour("C")
        sc_bond = self.sulphur.get_bond(carbon)

        structure = combine_structures([domain, self.structure])
        structure.break_bond(sh_bond)
        structure.break_bond(sc_bond)

        structure.make_bond(hydrogen, self.sulphur, structure.find_next_bond_nr())
        structure.make_bond(carbon, sulphur_1, structure.find_next_bond_nr())

        split = structure.split_disconnected_structures()

        tethered_polyketide = None

        for structure in split:
            if carbon in structure.graph:
                tethered_polyketide = structure
                break

        assert tethered_polyketide
        initialise_atom_attributes(tethered_polyketide)

        return tethered_polyketide


class PksElongationUnit:
    def __init__(self, name, smiles, c_idx, carbonyl_idx):
        """
        name: str, elongation unit name
        smiles: str, elongation unit SMILES
        c_idx: int, index of carbon that will be attached to the thioester carbon of the chain intermediate
        carbonyl_idx: int, index of the carbon of the carbonyl group
        """
        self.name = name
        self.smiles = smiles
        self.structure = read_smiles(self.smiles)
        initialise_atom_attributes(self.structure)
        self.structure.refresh_structure()

        self.c_to_pk_intermediate = None
        self.c_to_s = None

        for atom in self.structure.graph:
            if atom.nr == c_idx:
                self.c_to_pk_intermediate = atom
            elif atom.nr == carbonyl_idx:
                self.c_to_s = atom

        assert self.c_to_pk_intermediate is not None and self.c_to_s is not None

        self.c_to_pk_intermediate.annotations.in_central_chain = True
        self.c_to_s.annotations.in_central_chain = True

        if self.smiles == "O=CCC":
            # TODO: CHECK CHIRALITY ACTUALLY WORKS!!!!
            self.c_to_pk_intermediate.chiral = "clockwise"


def make_starter_monomer(name, smiles):
    assert name in _STARTER_PKS_TO_BACKBONE

    monomer = PksStarterUnit(name, smiles, _STARTER_PKS_TO_BACKBONE[name])
    return monomer


def make_elongation_monomer(name):
    if name == "MALONYL_COA":
        monomer = PksElongationUnit("Malonyl CoA", "CC=O", 0, 1)
    elif name == "METHYLMALONYL_COA":
        monomer = PksElongationUnit("Methylmalonyl CoA", "O=CCC", 2, 1)
    elif name == "METHOXYMALONYL_COA":
        monomer = PksElongationUnit("Methoxymalonyl CoA", "O=CCOC", 2, 1)
    elif name == "ETHYLMALONYL_COA":
        monomer = PksElongationUnit("Ethylmalonyl CoA", "O=CCCC", 2, 1)
    elif name == "WILDCARD":
        monomer = PksElongationUnit("Wildcard", "O=CC*", 2, 1)
    else:
        raise ValueError(f"{name} not recognised by RAIChU as PKS elongation unit.")

    return monomer


ESTER_BOND = BondDefiner("thioester_bond", "COC=O", 0, 1)
THIOESTERBOND = BondDefiner("thioester_bond", "SC(C)=O", 0, 1)
THIOESTERBOND_OXYGEN_INSERTED = BondDefiner(
    "thioester_bond_oxygen_inserted", "SC(O)=O", 0, 1
)
THIOESTERBOND_REDUCED_O = BondDefiner("thioester_bond_reduced_O", "SC(C)O", 0, 1)
THIOESTER_CARBON = GroupDefiner("thioester carbon", "SC(C)=O", 1)
THIOESTER_CARBON_REDUCED_O = GroupDefiner(
    "thioester carbon reduced oxygen", "SC(C)O", 1
)
THIOESTER_CARBON_OXYGEN_INSERTED = GroupDefiner(
    "thioester carbon oxygen inserted", "SC(O)=O", 1
)
LEAVING_OH_BOND = BondDefiner("Leaving -OH group bond", "C(=O)(O)CN", 0, 2)
N_AMINO_ACID = GroupDefiner("Nitrogen atom amino acid", "NCC(=O)O", 0)
C1_AMINO_ACID = GroupDefiner("C1 atom amino acid", "NCC(=O)O", 1)
C2_AMINO_ACID = GroupDefiner("C2 atom amino acid", "NCC(=O)O", 2)
N_AMINO_ACID_ATTACHED = GroupDefiner("Nitrogen atom amino acid", "NCC=O", 0)
C1_AMINO_ACID_ATTACHED = GroupDefiner("C1 atom amino acid", "NCC=O", 1)
C2_AMINO_ACID_ATTACHED = GroupDefiner("C2 atom amino acid", "NCC=O", 2)
OH_AMINO_ACID = GroupDefiner("OH atom amino acid", "NCC(O)=O", 3)
AMINO_ACID_BACKBONE = read_smiles("NCC(=O)O")
AMINO_ACID_BACKBONE_ATTACHED = read_smiles("NCC=O")
BETA_AMINO_ACID_BACKBONE = read_smiles("NCCC(=O)O")
B_N_AMINO_ACID = GroupDefiner("Nitrogen atom beta amino acid", "NCCC(=O)O", 0)
B_C1_AMINO_ACID = GroupDefiner("C1 atom beta amino acid", "NCCC(=O)O", 1)
B_C2_AMINO_ACID = GroupDefiner("C2 atom beta amino acid", "NCCC(=O)O", 2)
B_C3_AMINO_ACID = GroupDefiner("C3 atom beta amino acid", "NCCC(=O)O", 3)
B_LEAVING_BOND = BondDefiner("beta leaving bond", "NCCC(=O)O", 3, 5)
AMINO_FATTY_ACID = read_smiles("CCCCCCCC(=O)O")
ACID_C1 = GroupDefiner("C1 atom (fatty) acid", "CC(=O)O", 0)
ACID_C2 = GroupDefiner("C2 atom (fatty) acid", "CC(=O)O", 1)
COABOND = BondDefiner("CoA_bond", "CC(NCCC(NCCSC)=O)=O", 8, 9)
CYCLIC_CYS_C1_ATTACHED = GroupDefiner("oxidation_target", r"SC(=O)C1CSC=N1", 3)
CYCLIC_CYS_C2_ATTACHED = GroupDefiner("oxidation_target", r"SC(=O)C1CSC=N1", 4)
CYCLIC_SER_C1_ATTACHED = GroupDefiner("oxidation_target", r"SC(=O)C1COC=N1", 3)
CYCLIC_SER_C2_ATTACHED = GroupDefiner("oxidation_target", r"SC(=O)C1COC=N1", 4)

PEPTIDE_BOND = BondDefiner("peptide_bond", "CC(=O)NC", 1, 3)
CC_DOUBLE_BOND = BondDefiner("cc_double_bond", "C=C", 0, 1)
CO_DOUBLE_BOND = BondDefiner("co_double_bond", "C=O", 0, 1)
CC_SINGLE_BOND = BondDefiner("cc_single_bond", "CC", 0, 1)
C_TERMINAL_OH = BondDefiner("c_terminal_oh", "NCC(O)=O", 2, 3)
SH_BOND = BondDefiner("recent_elongation", "SC(C)=O", 0, 1)
SC_BOND = BondDefiner("recent_elongation", "SC(C)=O", 0, 1)
SC_BOND_MINIMAL = BondDefiner("minimal_sc_bond", "SC=O", 0, 1)
CO_BOND = BondDefiner("recent_elongation", "CO", 0, 1)
N_AMINO = GroupDefiner("N_amino", "CN", 1)
C_CARBOXYL = GroupDefiner("c_carboxyl", "C(O)=O", 0)
O_OH = GroupDefiner("O_oh", "CO", 1)
C_CH = GroupDefiner("C_Ch", "CC", 0)
N_NH = GroupDefiner("N_nh", "CN", 1)
KETO_GROUP = BondDefiner("keto_group", "C=O", 0, 1)
KETO_GROUP_C = GroupDefiner("keto_group_c", "C=O", 0)
O_BETAPROPRIOLACTONE = GroupDefiner("o_betapropriolactone", "SC(CCO)=O", 4)
O_BETAPROPRIOLACTONE_O = GroupDefiner("o_betapropriolactone", "OC(CCO)=O", 4)
O_BETAPROPRIOLACTONE_KETO_C = GroupDefiner("o_betapropriolactone_keto", "C(CCO)=O", 1)
O_BETAPROPRIOLACTONE_KETO_OH = GroupDefiner("o_betapropriolactone_keto", "C(CCO)=O", 4)
O_BETAPROPRIOLACTONE_TERMINAL_O = GroupDefiner("o_betapropriolactone", "OC(CCO)=O", 0)
O_BETAPROPRIOLACTONE_TERMINAL_S = GroupDefiner("o_betapropriolactone", "SC(CCO)=O", 0)
ASPARTIC_ACID = GroupDefiner("aspartic_acid", "NC(CC(O)=O)C=O", 4)
GLUTAMIC_ACID = GroupDefiner("glutamic_acid", "NC(CCC(O)=O)C=O", 5)
ARGININE_SECONDARY_N = GroupDefiner("arginine_secondary_n", "NC(CCCNC(N)=N)C(O)=O", 5)
CYSTEINE = GroupDefiner("cysteine", "NC(CS)C=O", 3)
SERINE = GroupDefiner("serine", "NC(CO)C=O", 3)
THREONINE = GroupDefiner("threonine", "CC(O)C(N)C=O", 2)
REDUCED_SERINE = GroupDefiner("reduced_serine", "NC(=C)C=O", 2)
REDUCED_THREONINE = GroupDefiner("reduced_threonine", "C=CC(N)C=O", 0)
PYROPHOSPHATE_BOND = BondDefiner("pyrophosphate_bond", "COP", 0, 1)
ATTACHED_SERINE_OX = GroupDefiner("attached_serine", "O=CNC(CO)C(S)=O", 0)
ATTACHED_CYSTEINE_OX = GroupDefiner("attached_cysteine", "O=CNC(CS)C(S)=O", 0)
ATTACHED_SERINE_O = GroupDefiner("attached_serine", "O=CNC(CO)C(S)=O", 5)
ATTACHED_CYSTEINE_S = GroupDefiner("attached_cysteine", "O=CNC(CS)C(S)=O", 5)
METHYLMALONYL_SIDECHAIN_CARBON = GroupDefiner("mmal_sidechain_c", "SC(=O)C(C)C(=O)", 4)
METHYLMALONYL_SULPHUR_CARBON = GroupDefiner("mmal_sidechain_sc", "SC(=O)C(C)C(=O)", 1)
METHYLMALONYL_MAINCHAIN_CARBON = GroupDefiner("mmal_sidechain_mc", "SC(=O)C(C)C(=O)", 5)
METHYLMALONYL_CHIRAL_CARBON = GroupDefiner("mmal_sidechain_mc", "SC(=O)C(C)C(=O)", 3)



from raichu.class_domain import *
from pikachu.reactions.functional_groups import find_atoms, GroupDefiner, combine_structures
from pikachu.reactions.basic_reactions import condensation
from raichu.reactions.general import initialise_atom_attributes
from pikachu.general import read_smiles

POLYKETIDE_S = GroupDefiner('Sulphur atom polyketide', 'SC(C)=O', 0)
POLYKETIDE_S_INSERTED_O = GroupDefiner(
    'Sulphur atom polyketide inserted O', 'SC(O)=O', 0)
POLYKETIDE_S_REDUCED_STARTER = GroupDefiner('Sulphur atom polyketide reduced O', 'SC(C)O', 0)
NRP_C = GroupDefiner('C atom to attach to PCP domain', 'NCC(O)=O', 2)
RIPP_N = GroupDefiner('N atom to attach to leader', 'NCC=O', 0)
AMINO_ACID_BACKBONE = read_smiles('NCC(=O)O')
AMINO_ACID_BACKBONE_N_TERMINUS = read_smiles('NCC(=O)N')
B_AMINO_ACID_BACKBONE = read_smiles('NCCC(=O)O')
ASP_ACID_C = GroupDefiner('Asp', 'NC(C=O)CC(O)=O', 5)
B_NRP_C = GroupDefiner('Beta C atom to attach to PCP domain', 'NCCC(O)=O', 3)
MAL_AMINO = GroupDefiner('malonyl_starter_amino', 'NC(=O)CC(O)=O', 4)


def condensation_with_reverse_numbering(structure_1, structure_2, oh_bond, h_bond):
    """
    Returns condensed product of structure 1 and structure 2, and a water molecule, where the numbering of structure 1 is kept

    Input
    ----------
    structure_1: Structure object, structure containing -OH leaving group
    structure_2: Structure object, structure containing -H leaving group
    oh_bond: Bond object, bond attached to -OH leaving group
    h_bond: Bond object, bond attached to -H leaving group

    Returns
    -------
    list of [product, water], both Structure objects, with product the condensed product of structure 1 and structure 2
        and water a water molecule.

    """

    # Ensure that the defined OH-bond actually has an -OH leaving group
    o_atom = None
    o_neighbour = None

    for atom in oh_bond.neighbours:
        if not o_atom and atom.type == 'O' and atom.has_neighbour('H'):
            o_atom = atom
        else:
            o_neighbour = atom

    if not o_atom:
        raise Exception(
            f"The selected bond {oh_bond} does not attach to an -OH leaving group.")

    # Ensure that the defined H-bond actually has an -H leaving group

    h_atom = None
    h_neighbour = None

    for atom in h_bond.neighbours:
        if not h_atom and atom.type == 'H':
            h_atom = atom
        else:
            h_neighbour = atom

    if not h_atom:
        raise Exception(
            f"The selected bond {h_bond} does not attach to an -H leaving group.")

    # Break the bonds between the leaving groups and the rest of the product

    structure_1.break_bond(oh_bond)
    structure_2.break_bond(h_bond)

    # Put the two structures into the same object
    structure = combine_structures([structure_2, structure_1])
    # Retrieve the atoms that have to form bonds: h_atom with o_atom to form water, and h_neighbour with o_neighbour
    # to form the product

    h_atom = structure.get_atom(h_atom)
    o_atom = structure.get_atom(o_atom)
    h_neighbour = structure.get_atom(h_neighbour)
    o_neighbour = structure.get_atom(o_neighbour)

    # Create the bonds

    structure.make_bond(h_atom, o_atom, structure.find_next_bond_nr())
    structure.make_bond(h_neighbour, o_neighbour,
                        structure.find_next_bond_nr())

    # Put the water and the product into different Structure instances

    structures = structure.split_disconnected_structures()

    water = None
    product = None

    # Find out which of the structures is your product and which is your water

    for structure in structures:
        if h_atom in structure.graph:
            water = structure
        elif h_neighbour in structure.graph:
            structure.refresh_structure(find_cycles=True)
            product = structure

    return [product, water]


def attach_to_domain_pk(polyketide):
    """
    Attaches the sulphur atom in the input polyketide to a PKS domain and
    returns the attached structure as a PIKAChU Structure object

    polyketide: PIKAChU Structure object, to-be attached structure
    """
    # Add attributes to input molecule
    initialise_atom_attributes(polyketide)

    # Create domain
    domain = make_scaffold_domain('ACP')
    sh_bond = domain.bond_lookup[domain.atoms[1]][domain.atoms[2]]
    hydrogen = domain.atoms[2]
    sulphur_1 = domain.atoms[1]

    locations_sulphur = find_atoms(POLYKETIDE_S, polyketide)
    if len(locations_sulphur) == 0:
        locations_sulphur = find_atoms(
            POLYKETIDE_S_INSERTED_O, polyketide)
        if len(locations_sulphur) == 0:
            locations_sulphur = find_atoms(POLYKETIDE_S_REDUCED_STARTER, polyketide)
    assert len(locations_sulphur) == 1
    sulphur_2 = locations_sulphur[0]
    carbon = sulphur_2.get_neighbour('C')
    sc_bond = sulphur_2.get_bond(carbon)

    structure = combine_structures([domain, polyketide])
    structure.break_bond(sh_bond)
    structure.break_bond(sc_bond)

    structure.make_bond(hydrogen, sulphur_2, structure.find_next_bond_nr())
    structure.make_bond(carbon, sulphur_1, structure.find_next_bond_nr())

    split = structure.split_disconnected_structures()

    tethered_polyketide = None

    for structure in split:
        if carbon in structure.graph:
            tethered_polyketide = structure
            break

    assert tethered_polyketide
    initialise_atom_attributes(tethered_polyketide)

    return tethered_polyketide


def attach_to_domain_nrp(nrp):
    """
    Attaches the input NRP to a PCP domain and returns the product as a
    PIKAChU Structure object

    nrp: PIKAChU Structure object, to-be attached NRP
    """

    initialise_atom_attributes(nrp)

    # Create domain
    domain = make_scaffold_domain('PCP')

    # Remove OH group from carboxylic acid in NRP, to allow attachment
    # to domain

    if nrp.find_substructures(AMINO_ACID_BACKBONE):
        locations_c_to_domain = find_atoms(NRP_C, nrp)
        assert len(locations_c_to_domain) == 1
        c_atom_to_domain = locations_c_to_domain[0]
    elif nrp.find_substructures(B_AMINO_ACID_BACKBONE):

        locations_c_to_domain = find_atoms(B_NRP_C, nrp)
        location_c_indices = []

        for i, c_atom in enumerate(locations_c_to_domain):
            for neighbour in c_atom.neighbours:
                if neighbour.annotations.leaving_oh_o:
                    location_c_indices.append(i)

        assert len(location_c_indices) == 1
        index = location_c_indices[0]
        c_atom_to_domain = locations_c_to_domain[index]

        # asp_acid_cs = find_atoms(ASP_ACID_C, nrp)
        # print(asp_acid_cs)
        # mal_amino_cs = find_atoms(MAL_AMINO, nrp)
        # print(mal_amino_cs)

        # for asp_acid_c in asp_acid_cs:
        #     if asp_acid_c in locations_c_to_domain:
        #         locations_c_to_domain.remove(asp_acid_c)
        # for mal_amino_c in mal_amino_cs:
        #     if mal_amino_c in locations_c_to_domain:
        #         locations_c_to_domain.remove(mal_amino_c)
        # assert len(locations_c_to_domain) == 1
        # c_atom_to_domain = locations_c_to_domain[0]
    else:
        c_atom_to_domain = None
        for atom in nrp.graph:
            if atom.annotations.c2_acid:
                c_atom_to_domain = atom

    assert c_atom_to_domain

    oxygens = c_atom_to_domain.get_neighbours('O')

    hydroxyl_oxygen = None
    hydroxyl_bond = None

    hydrogen_bond = domain.bond_lookup[domain.atoms[1]][domain.atoms[2]]

    assert hydrogen_bond.has_neighbour('S')
    assert hydrogen_bond.has_neighbour('H')

    for oxygen in oxygens:
        bond = c_atom_to_domain.get_bond(oxygen)
        if bond.type == 'single' and oxygen.has_neighbour('H'):
            hydroxyl_oxygen = oxygen
            hydroxyl_bond = bond
            break

    assert hydroxyl_oxygen and hydroxyl_bond

    structure = condensation(nrp, domain, hydroxyl_bond, hydrogen_bond)[0]

    initialise_atom_attributes(structure)

    return structure


def attach_to_follower_ripp(ripp):
    initialise_atom_attributes(ripp)

    # Create domain
    domain = make_scaffold_peptide('Follower')
    # Remove OH group from carboxylic acid in NRP, to allow attachment
    # to domain

    if ripp.find_substructures(AMINO_ACID_BACKBONE):
        locations_c_to_domain = find_atoms(NRP_C, ripp)
        assert len(locations_c_to_domain) == 1
        c_atom_to_domain = locations_c_to_domain[0]
    elif ripp.find_substructures(B_AMINO_ACID_BACKBONE):
        locations_c_to_domain = find_atoms(B_NRP_C, ripp)
        asp_acid_cs = find_atoms(ASP_ACID_C, ripp)
        mal_amino_cs = find_atoms(MAL_AMINO, ripp)

        for asp_acid_c in asp_acid_cs:
            if asp_acid_c in locations_c_to_domain:
                locations_c_to_domain.remove(asp_acid_c)
        for mal_amino_c in mal_amino_cs:
            if mal_amino_c in locations_c_to_domain:
                locations_c_to_domain.remove(mal_amino_c)
        assert len(locations_c_to_domain) == 1
        c_atom_to_domain = locations_c_to_domain[0]
    else:
        c_atom_to_domain = None
        for atom in ripp.graph:
            if atom.annotations.c2_acid:
                c_atom_to_domain = atom

    assert c_atom_to_domain

    oxygens = c_atom_to_domain.get_neighbours('O')

    hydroxyl_oxygen = None
    hydroxyl_bond = None

    hydrogen_bond = domain.bond_lookup[domain.atoms[1]][domain.atoms[2]]

    assert hydrogen_bond.has_neighbour('N')
    assert hydrogen_bond.has_neighbour('H')

    for oxygen in oxygens:
        bond = c_atom_to_domain.get_bond(oxygen)
        if bond.type == 'single' and oxygen.has_neighbour('H'):
            hydroxyl_oxygen = oxygen
            hydroxyl_bond = bond
            break

    assert hydroxyl_oxygen and hydroxyl_bond
    structure = condensation_with_reverse_numbering(ripp, domain, hydroxyl_bond, hydrogen_bond)[0]
    initialise_atom_attributes(structure)
    return structure


def attach_to_leader_ripp(ripp):
    initialise_atom_attributes(ripp)

    # Create domain
    domain = make_scaffold_peptide('Leader')
    n_atom_to_domain = None
    if ripp.find_substructures(AMINO_ACID_BACKBONE_N_TERMINUS):
        locations_n_to_domain = find_atoms(RIPP_N, ripp)
        for nitrogen in locations_n_to_domain:
            if len(nitrogen.get_neighbours('H')) == 2:
                n_atom_to_domain = nitrogen
                break
    assert n_atom_to_domain
    oxygen_bond = domain.bond_lookup[domain.atoms[0]][domain.atoms[1]]

    assert oxygen_bond.has_neighbour('O')
    hydrogen = n_atom_to_domain.get_neighbours('H')[0]
    nitrogen_bond = n_atom_to_domain.get_bond(hydrogen)
    assert nitrogen_bond
    structure = condensation(
        domain, ripp, oxygen_bond, nitrogen_bond)[0]
    initialise_atom_attributes(structure)

    return structure



from enum import Enum, unique
import itertools

from pikachu.drawing.drawing import Drawer
from raichu.reactions.general_tailoring_reactions import (
    proteolytic_cleavage,
    find_atoms_for_tailoring,
    remove_atom,
    single_bond_oxidation,
    addition,
    oxidative_bond_formation,
    epoxidation,
    double_bond_reduction,
    double_bond_shift,
    macrolactam_formation,
    cyclodehydration,
    change_chirality,
    excise_from_structure,
    reductive_bond_breakage,
)
from raichu.data.attributes import PRENYL_TRANSFERASE_SUBSTRATES_TO_SMILES
from raichu.data.molecular_moieties import (
    CO_BOND,
    CC_DOUBLE_BOND,
    PEPTIDE_BOND,
    CC_SINGLE_BOND,
    KETO_GROUP,
    C_CARBOXYL,
    ASPARTIC_ACID,
    GLUTAMIC_ACID,
    CYSTEINE,
    SERINE,
    THREONINE,
    REDUCED_SERINE,
    REDUCED_THREONINE,
    C1_AMINO_ACID_ATTACHED,
    ARGININE_SECONDARY_N,
    ESTER_BOND,
)
from pikachu.reactions.functional_groups import find_atoms, find_bonds
from pikachu.reactions.basic_reactions import hydrolysis


@unique
class TailoringEnzymeType(Enum):
    # Group transfer reactions
    METHYLTRANSFERASE = 1
    C_METHYLTRANSFERASE = 2
    N_METHYLTRANSFERASE = 3
    O_METHYLTRANSFERASE = 4
    HYDROXYLASE = 5
    EPOXIDASE = 6
    PRENYLTRANSFERASE = 7
    ACETYLTRANSFERASE = 8
    ACYLTRANSFERASE = 9
    AMINOTRANSFERASE = 10
    HALOGENASE = 11
    METHYL_MUTASE = 12

    # Oxidoreduction
    DOUBLE_BOND_REDUCTASE = 13
    DOUBLE_BOND_ISOMERASE = 14
    DEHYDROGENASE = 15
    KETO_REDUCTION = 16
    ALCOHOL_DEHYDROGENASE = 17

    # Elimination
    PEPTIDASE = 18
    PROTEASE = 19
    MONOAMINE_OXIDASE = 20
    DEHYDRATASE = 21
    THREONINE_SERINE_DEHYDRATASE = 22
    DECARBOXYLASE = 23
    SPLICEASE = 24
    ARGINASE = 25

    # Cyclization
    OXIDATIVE_BOND_SYNTHASE = 26
    MACROLACTAM_SYNTHETASE = 27
    CYCLODEHYDRASE = 28
    LANTHIPEPTIDE_CYCLASE = 29
    LANTHIONINE_SYNTHETASE = 30
    THIOPEPTIDE_CYCLASE = 31

    # Epimerization
    AMINO_ACID_EPIMERASE = 32

    # Bond breakage
    HYDROLASE = 33
    REDUCTIVE_LYASE = 34

    @staticmethod
    def from_string(label: str) -> "TailoringEnzymeType":
        for value in TailoringEnzymeType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown tailoring enzyme: {label}")


class TailoringEnzyme:
    def __init__(
        self,
        gene_name,
        enzyme_type,
        modification_sites: list = None,
        substrate: str = None,
    ) -> None:
        self.gene_name = gene_name
        self.type = TailoringEnzymeType.from_string(enzyme_type)
        self.modification_sites = modification_sites
        self.substrate = substrate

    def do_tailoring(self, structure):
        """
        Performs tailoring reaction
        """
        if len(self.modification_sites) == 0:
            return structure
        if self.type.name == "HYDROXYLASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom = atom[0]  # only one atom is hydroxylated at a time
                atom = structure.get_atom(atom)
                structure = addition(atom, "O", structure)
        elif self.type.name in [
            "METHYLTRANSFERASE",
            "C_METHYLTRANSFERASE",
            "N_METHYLTRANSFERASE",
            "O_METHYLTRANSFERASE",
        ]:
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom = atom[0]  # only one atom is methylated at a time
                atom = structure.get_atom(atom)
                structure = addition(atom, "C", structure)
        elif self.type.name == "PRENYLTRANSFERASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom = atom[0]  # only one atom is methylated at a time
                atom = structure.get_atom(atom)
                if self.substrate not in PRENYL_TRANSFERASE_SUBSTRATES_TO_SMILES:
                    raise ValueError(
                        f"Not implemented prenyltransferase substrate: {self.substrate}"
                    )
                substrate = PRENYL_TRANSFERASE_SUBSTRATES_TO_SMILES[self.substrate]
                structure = addition(atom, substrate, structure)
        elif self.type.name == "ACETYLTRANSFERASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom = atom[0]  # only one atom is methylated at a time
                atom = structure.get_atom(atom)
                structure = addition(atom, "[H]C(C)=O", structure)
        elif self.type.name == "ACYLTRANSFERASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom = atom[0]  # only one atom is methylated at a time
                atom = structure.get_atom(atom)
                if self.substrate:
                    structure = addition(atom, self.substrate, structure)
        elif self.type.name == "OXIDATIVE_BOND_SYNTHASE":
            for atoms in self.modification_sites:
                if len(atoms) < 2:
                    continue
                atom1 = structure.get_atom(atoms[0])
                carbon_1 = structure.get_atom(atoms[1])
                structure = oxidative_bond_formation(atom1, carbon_1, structure)
        elif self.type.name == "EPOXIDASE":
            for atoms in self.modification_sites:
                if len(atoms) < 2:
                    continue
                atom1 = structure.get_atom(atoms[0])
                carbon_1 = structure.get_atom(atoms[1])
                structure = epoxidation(atom1, carbon_1, structure)
        elif self.type.name == "DOUBLE_BOND_REDUCTASE":
            for atoms in self.modification_sites:
                if len(atoms) < 2:
                    continue
                atom1 = structure.get_atom(atoms[0])
                carbon_1 = structure.get_atom(atoms[1])
                structure = double_bond_reduction(atom1, carbon_1, structure)
        elif self.type.name == "DEHYDROGENASE":
            for atoms in self.modification_sites:
                if len(atoms) < 2:
                    continue
                atom1 = structure.get_atom(atoms[0])
                carbon_1 = structure.get_atom(atoms[1])
                structure = single_bond_oxidation(atom1, carbon_1, structure)
        elif self.type.name == "DOUBLE_BOND_ISOMERASE":
            for atoms in self.modification_sites:
                if len(atoms) < 4:
                    continue
                old_double_bond_atom1 = structure.get_atom(atoms[0])
                old_double_bond_atom2 = structure.get_atom(atoms[1])
                new_double_bond_atom1 = structure.get_atom(atoms[2])
                new_double_bond_atom2 = structure.get_atom(atoms[3])
                if len(set(atoms)) == len(atoms):
                    raise ValueError(
                        "The bonds need to be adjacent to perform a dauble bond shift."
                    )
                structure = double_bond_shift(
                    structure,
                    old_double_bond_atom1,
                    old_double_bond_atom2,
                    new_double_bond_atom1,
                    new_double_bond_atom2,
                )
        elif self.type.name == "AMINOTRANSFERASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                oxygen = atom1.get_neighbour("O")
                structure = double_bond_reduction(atom1, oxygen, structure)
                oxygen = structure.get_atom(oxygen)
                structure = remove_atom(oxygen, structure)
                atom = structure.get_atom(atom)
                structure = addition(atom, "N", structure)
        elif self.type.name == "KETO_REDUCTION":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                if atom1.type != "O":
                    raise ValueError(
                        f"Can not perform KETO_REDUCTION on atom {atom1}, since there is no oxygen to be reduced."
                    )
                carbon_1 = atom1.get_neighbour("C")
                structure = double_bond_reduction(atom1, carbon_1, structure)
        elif self.type.name == "ALCOHOL_DEHYDROGENASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                if atom1.type != "O":
                    raise ValueError(
                        f"Can not perform ALCOHOL_DEHYDROGENASE on atom {atom1}, since there is no oxygen to be reduced."
                    )
                carbon_1 = atom1.get_neighbour("C")
                structure = single_bond_oxidation(atom1, carbon_1, structure)
        elif self.type.name == "DECARBOXYLASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                structure = remove_atom(atom1, structure)
        elif self.type.name == "DEHYDRATASE":
            for atoms in self.modification_sites:
                if len(atoms) < 2:
                    continue
                atom1 = structure.get_atom(atoms[0])
                carbon_1 = structure.get_atom(atoms[1])
                oxygen = atom1.get_neighbour("O")
                if not oxygen:
                    oxygen = carbon_1.get_neighbour("O")
                if not oxygen:
                    raise ValueError(
                        f"Can not perform DEHYDRATASE on atoms {atom1} and {carbon_1}, since there is no hydroxygroup to be removed."
                    )
                structure = remove_atom(oxygen, structure)
                if atom1.get_bond(carbon_1):
                    structure = single_bond_oxidation(atom1, carbon_1, structure)
                else:
                    structure = oxidative_bond_formation(atom1, carbon_1, structure)
        elif self.type.name == "MONOAMINE_OXIDASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                if atom1.type != "N":
                    raise ValueError(
                        f"Can not perform MONOAMINE_OXYDASE on atom {atom1}, since there is no nitrogen to be removed."
                    )
                structure = remove_atom(atom1, structure)
        elif self.type.name == "HALOGENASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                if self.substrate in ["F", "Cl", "Br", "I"]:
                    structure = addition(atom1, self.substrate, structure)
        elif self.type.name in ["PROTEASE", "PEPTIDASE"]:
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                atom1 = structure.get_atom(atoms[0])
                carbon_1 = structure.get_atom(atoms[1])
                structure = proteolytic_cleavage(atom1.get_bond(carbon_1), structure)
        elif self.type.name == "MACROLACTAM_SYNTHETASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                structure = macrolactam_formation(structure, atom1)
        elif self.type.name == "CYCLODEHYDRASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                break_all = False
                for neighbour_1 in atom1.neighbours:
                    if neighbour_1.type == "C":
                        for neighbour_2 in neighbour_1.neighbours:
                            if neighbour_2.type == "C":
                                nitrogen = neighbour_2.get_neighbour("N")
                                if nitrogen:
                                    carbon = nitrogen.get_neighbour("C")
                                    oxygen = carbon.get_neighbour("O")
                                    if carbon and nitrogen:
                                        structure = cyclodehydration(
                                            structure, atom1, oxygen
                                        )
                                        # breaking out of all loops
                                        break_all = True
                                        break
                        if break_all:
                            break
                else:
                    raise ValueError(
                        "No downstream amino acid for cyclodehydration availiable."
                    )
        elif self.type.name == "THIOPEPTIDE_CYCLASE":
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                carbon_1 = structure.get_atom(atoms[0])
                carbon_2_candidates = carbon_1.get_neighbours("C")
                carbon_2 = None
                for carbon_2_candidate in carbon_2_candidates:
                    if carbon_2_candidate.has_neighbour("N"):
                        carbon_2 = carbon_2_candidate
                assert carbon_2
                carbon_3 = structure.get_atom(atoms[1])
                carbon_4_candidates = carbon_3.get_neighbours("C")
                carbon_4 = None
                for carbon_4_candidate in carbon_4_candidates:
                    if carbon_4_candidate.has_neighbour("N"):
                        carbon_4 = carbon_4_candidate
                assert carbon_4
                nitrogen = carbon_4.get_neighbour("N")
                carbon_5 = [
                    carbon
                    for carbon in nitrogen.get_neighbours("C")
                    if carbon != carbon_4
                ][0]
                assert nitrogen, carbon_5
                oxygen = carbon_5.get_neighbour("O")
                structure = double_bond_reduction(carbon_1, carbon_2, structure)
                structure = double_bond_reduction(carbon_5, oxygen, structure)
                structure = double_bond_shift(
                    structure, carbon_3, carbon_4, carbon_4, nitrogen
                )
                structure = oxidative_bond_formation(carbon_1, carbon_3, structure)
                structure = oxidative_bond_formation(carbon_2, carbon_5, structure)
        elif self.type.name == "THREONINE_SERINE_DEHYDRATASE":
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                atom1 = atom[0]  # only one atom is modified at a time
                atom1 = structure.get_atom(atom1)
                carbon_1 = atom1.get_neighbour("C")
                carbon_2_candidates = carbon_1.get_neighbours("C")
                carbon_2 = None
                for carbon_2_candidate in carbon_2_candidates:
                    if carbon_2_candidate.has_neighbour("N"):
                        carbon_2 = carbon_2_candidate
                assert carbon_2
                structure = remove_atom(atom1, structure)
                structure = single_bond_oxidation(carbon_2, carbon_1, structure)
        elif self.type.name == "LANTHIPEPTIDE_CYCLASE":
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                if structure.get_atom(atoms[0]).type == "S":
                    sulfur = structure.get_atom(atoms[0])
                    carbon_1 = structure.get_atom(atoms[1])
                    carbon_2_candidates = carbon_1.get_neighbours("C")
                    carbon_2 = None
                    for carbon_2_candidate in carbon_2_candidates:
                        if carbon_2_candidate.has_neighbour("N"):
                            carbon_2 = carbon_2_candidate
                    assert carbon_2
                    structure = double_bond_reduction(carbon_1, carbon_2, structure)
                    structure = oxidative_bond_formation(sulfur, carbon_1, structure)
                if structure.get_atom(atoms[0]).type == "C":
                    carbon_1_1 = structure.get_atom(atoms[0])
                    carbon_1_2_candidates = carbon_1_1.get_neighbours("C")
                    carbon_1_2 = None
                    for carbon_1_2_candidate in carbon_1_2_candidates:
                        if carbon_1_2_candidate.has_neighbour("N"):
                            carbon_1_2 = carbon_1_2_candidate
                    assert carbon_1_2
                    carbon_2_1 = structure.get_atom(atoms[1])  # already cyclized carbon
                    carbon_2_2_candidates = carbon_2_1.get_neighbours("C")
                    carbon_2_2 = None
                    for carbon_2_2_candidate in carbon_2_2_candidates:
                        if carbon_2_2_candidate.has_neighbour("N"):
                            carbon_2_2 = carbon_2_2_candidate
                    assert carbon_2_2
                    if carbon_2_1.get_bond(carbon_2_2).type == "double":
                        structure = double_bond_reduction(
                            carbon_2_1, carbon_2_2, structure
                        )
                    structure = double_bond_reduction(carbon_1_1, carbon_1_2, structure)
                    structure = oxidative_bond_formation(
                        carbon_1_1, carbon_2_2, structure
                    )
        elif self.type.name == "LANTHIONINE_SYNTHETASE":
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                if structure.get_atom(atoms[0]).type == "S":
                    sulfur = structure.get_atom(atoms[0])
                    carbon_1 = structure.get_atom(atoms[1])
                    oxygen = carbon_1.get_neighbour("O")
                    if oxygen:
                        structure = remove_atom(oxygen, structure)
                    structure = oxidative_bond_formation(sulfur, carbon_1, structure)
                if structure.get_atom(atoms[0]).type == "C":
                    carbon = structure.get_atom(atoms[0])
                    oxygen_2 = carbon.get_neighbour("O")
                    carbon_1 = structure.get_atom(atoms[1])
                    oxygen = carbon_1.get_neighbour("O")
                    carbon_2_candidates = carbon_1.get_neighbours("C")
                    carbon_2 = None
                    for carbon_2_candidate in carbon_2_candidates:
                        if carbon_2_candidate.has_neighbour("N"):
                            carbon_2 = carbon_2_candidate
                    assert carbon_2
                    if oxygen:
                        structure = remove_atom(oxygen, structure)
                    if oxygen_2:
                        structure = remove_atom(oxygen_2, structure)
                    structure = oxidative_bond_formation(carbon, carbon_2, structure)
        elif self.type.name == "AMINO_ACID_EPIMERASE":
            # atom needs to be amino acid alpha atom
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                carbon = structure.get_atom(atom[0])
                assert carbon.type == "C"
                structure = change_chirality(carbon, structure)

        elif self.type.name == "SPLICEASE":
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                carbon_1 = structure.get_atom(atoms[0])
                carbon_2 = structure.get_atom(atoms[1])
                structure = excise_from_structure(carbon_1, carbon_2, structure)
        elif self.type.name == "HYDROLASE":
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                hetero_atom = structure.get_atom(atoms[0])
                carbon = structure.get_atom(atoms[1])  # gets oh attached
                bond = hetero_atom.get_bond(carbon)
                structure = hydrolysis(structure, bond)
        elif self.type.name == "REDUCTIVE_LYASE":
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                carbon_1 = structure.get_atom(atoms[0])
                carbon_2 = structure.get_atom(atoms[1])
                structure = reductive_bond_breakage(carbon_1, carbon_2, structure)
        elif self.type.name == "ARGINASE":
            # atom needs to be arginine secondary nitrogen
            for atom in self.modification_sites:
                if len(atom) == 0:
                    continue
                nitrogen = structure.get_atom(atom[0])
                assert nitrogen.type == "N"
                carbon = [
                    carbon
                    for carbon in nitrogen.get_neighbours("C")
                    if [atom.type for atom in carbon.neighbours].count("N") == 3
                ][0]
                bond = nitrogen.get_bond(carbon)
                assert bond
                structure.break_bond(bond)
                structure_1, structure_2 = structure.split_disconnected_structures()
                if nitrogen in structure_1.graph:
                    structure = structure_1
                else:
                    structure = structure_2
                structure.add_atom("H", [nitrogen])
                structure.refresh_structure(find_cycles=True)
        elif self.type.name == "METHYL_MUTASE":
            for atoms in self.modification_sites:
                if len(atoms) != 2:
                    continue
                transferred_c = atoms[0]
                # Assert its actually a methyl group
                assert [atom.type for atom in transferred_c.neighbours].count("H") == 3
                source = None
                source = [
                    atom for atom in transferred_c.neighbours if atom.type != "H"
                ][0]
                assert source
                destination_c = atoms[1]
                assert destination_c.has_neighbour("H")
                assert transferred_c.type == "C" and destination_c.type == "C"

                structure = reductive_bond_breakage(source, transferred_c, structure)
                structure = addition(destination_c, "C", structure)
                structure.refresh_structure(find_cycles=True)

        return structure

    def get_possible_sites(self, structure, out_file=None):
        possible_sites = []
        if self.type.name in [
            "HYDROXYLASE",
        ]:
            possible_sites.extend(
                [
                    [atom]
                    for atom in find_atoms_for_tailoring(structure, "C")
                    if atom.has_neighbour("H")
                ]
            )
        elif self.type.name in [
            "C_METHYLTRANSFERASE",
            "N_METHYLTRANSFERASE",
            "O_METHYLTRANSFERASE",
        ]:
            atom = self.type.name.split("_")[0]
            possible_sites.extend(
                [
                    [atom]
                    for atom in find_atoms_for_tailoring(structure, atom)
                    if atom.has_neighbour("H")
                ]
            )
        elif self.type.name in [
            "METHYLTRANSFERASE",
            "PRENYLTRANSFERASE",
            "ACETYLTRANSFERASE",
            "ACYLTRANSFERASE",
            "OXIDATIVE_BOND_SYNTHASE",
            "HALOGENASE",
            "METHYL_MUTASE",
        ]:
            possible_sites.extend(
                [
                    [atom]
                    for atom in find_atoms_for_tailoring(structure, "C")
                    if atom.has_neighbour("H")
                ]
            )
            possible_sites.extend(
                [
                    [atom]
                    for atom in find_atoms_for_tailoring(structure, "N")
                    if atom.has_neighbour("H")
                ]
            )
            possible_sites.extend(
                [
                    [atom]
                    for atom in find_atoms_for_tailoring(structure, "O")
                    if atom.has_neighbour("H")
                ]
            )
            possible_sites.extend(
                [
                    [atom]
                    for atom in find_atoms_for_tailoring(structure, "S")
                    if atom.has_neighbour("H")
                ]
            )
        elif self.type.name in ["SPLICEASE"]:
            possible_sites.extend(
                [[atom] for atom in find_atoms_for_tailoring(structure, "C")]
            )
            possible_sites.extend(
                [[atom] for atom in find_atoms_for_tailoring(structure, "N")]
            )
            possible_sites.extend(
                [[atom] for atom in find_atoms_for_tailoring(structure, "O")]
            )
            possible_sites.extend(
                [[atom] for atom in find_atoms_for_tailoring(structure, "S")]
            )

        elif self.type.name in ["EPOXIDASE", "DOUBLE_BOND_REDUCTASE"]:
            peptide_bonds = find_bonds(CC_DOUBLE_BOND, structure)
            for bond in peptide_bonds:
                possible_sites.append(bond.neighbours)

        elif self.type.name == "DEHYDROGENASE":
            peptide_bonds = find_bonds(CC_SINGLE_BOND, structure)
            for bond in peptide_bonds:
                possible_sites.append(bond.neighbours)

        elif self.type.name == "REDUCTIVE_LYASE":
            possible_sites.extend(
                [
                    bond.neighbours
                    for bond in structure.bonds.values()
                    if "H" not in [atom.type for atom in bond.neighbours]
                ]
            )

        elif self.type.name == "DOUBLE_BOND_ISOMERASE":
            peptide_bonds = find_bonds(CC_DOUBLE_BOND, structure)
            for bond in peptide_bonds:
                neighbouring_bonds = bond.get_neighbouring_bonds()
                for neighbouring_bond in neighbouring_bonds:
                    if not "H" in [atom.type for atom in neighbouring_bond.neighbours]:
                        possible_sites.append(
                            bond.neighbours + neighbouring_bond.neighbours
                        )

        elif self.type.name == "AMINOTRANSFERASE":
            possible_sites.extend(
                [[atom] for atom in find_atoms(KETO_GROUP, structure)]
            )

        elif self.type.name == "KETO_REDUCTION":
            oxygens = find_atoms(KETO_GROUP, structure)
            possible_sites.extend([[atom] for atom in oxygens])

        elif self.type.name == "ALCOHOL_DEHYDROGENASE":
            possible_sites.extend(
                [[atom] for atom in find_atoms_for_tailoring(structure, "O")]
            )

        elif self.type.name == "DECARBOXYLASE":
            possible_sites.extend(
                [[atom] for atom in find_atoms(C_CARBOXYL, structure)]
            )

        elif self.type.name == "DEHYDRATASE":
            co_bonds = find_bonds(CO_BOND, structure)
            for co_bond in co_bonds:
                neighbouring_bonds = co_bond.get_neighbouring_bonds()
                for neighbouring_bond in neighbouring_bonds:
                    if (
                        not "H" in [atom.type for atom in neighbouring_bond.neighbours]
                        and neighbouring_bond.type == "single"
                    ):
                        for neighbouring_atom in neighbouring_bond.neighbours:
                            if neighbouring_atom != co_bond.get_neighbour(
                                "C"
                            ) and neighbouring_atom.has_neighbour("H"):
                                possible_sites.append(neighbouring_bond.neighbours)

        elif self.type.name == "MONOAMINE_OXIDASE":
            n_atoms_with_one_h = find_atoms_for_tailoring(structure, "N")
            for n_atom in n_atoms_with_one_h:
                if [atom.type for atom in n_atom.neighbours].count("H") == 2:
                    possible_sites.append([n_atom])
        elif self.type.name in ["PROTEASE", "PEPTIDASE"]:
            peptide_bonds = find_bonds(PEPTIDE_BOND, structure)
            for bond in peptide_bonds:
                possible_sites.append(bond.neighbours)
        elif self.type.name in ["HYDROLASE"]:
            ester_bonds = find_bonds(ESTER_BOND, structure)
            for bond in ester_bonds:
                possible_sites.append(bond.neighbours)
        elif self.type.name == "MACROLACTAM_SYNTHETASE":
            asp_glu_oxygen = find_atoms(ASPARTIC_ACID, structure) + find_atoms(
                GLUTAMIC_ACID, structure
            )
            possible_sites.extend(asp_glu_oxygen)
        elif self.type.name == "CYCLODEHYDRASE":
            cys_ser_thr_x = (
                find_atoms(CYSTEINE, structure)
                + find_atoms(SERINE, structure)
                + find_atoms(THREONINE, structure)
            )
            possible_sites.extend([[atom] for atom in cys_ser_thr_x])
        elif self.type.name == "THREONINE_SERINE_DEHYDRATASE":
            ser_thr_x = find_atoms(SERINE, structure) + find_atoms(THREONINE, structure)
            possible_sites.extend([[atom] for atom in ser_thr_x])
        elif self.type.name == "LANTHIPEPTIDE_CYCLASE":
            cys_x = find_atoms(CYSTEINE, structure)
            ser_thr_c = find_atoms(REDUCED_SERINE, structure) + find_atoms(
                REDUCED_THREONINE, structure
            )
            combinations = [list(t) for t in itertools.product(cys_x, ser_thr_c)]
            combinations.extend(
                [list(t) for t in itertools.product(ser_thr_c, ser_thr_c)]
            )
            possible_sites.extend(combinations)
        elif self.type.name == "LANTHIONINE_SYNTHETASE":
            cys_x = find_atoms(CYSTEINE, structure)
            ser_thr_x = find_atoms(SERINE, structure) + find_atoms(THREONINE, structure)
            ser_thr_c = [atom.get_neighbour("C") for atom in ser_thr_x]
            combinations = [list(t) for t in itertools.product(cys_x, ser_thr_c)]
            combinations.extend(
                [list(t) for t in itertools.product(ser_thr_c, ser_thr_c)]
            )
            possible_sites.extend(combinations)
        elif self.type.name == "AMINO_ACID_EPIMERASE":
            alpha_cs_amino_acid_backbone = find_atoms(C1_AMINO_ACID_ATTACHED, structure)
            possible_sites.extend([[atom] for atom in alpha_cs_amino_acid_backbone])
        elif self.type.name == "ARGINASE":
            arginine_n = find_atoms(ARGININE_SECONDARY_N, structure)
            possible_sites.extend([[atom] for atom in arginine_n])
        elif self.type.name == "THIOPEPTIDE_CYCLASE":
            ser_thr_c = find_atoms(REDUCED_SERINE, structure) + find_atoms(
                REDUCED_THREONINE, structure
            )
            possible_sites.extend(
                [list(t) for t in itertools.product(ser_thr_c, ser_thr_c)]
            )

        if out_file:
            drawing = Drawer(structure)
            site_list = []

            for possible_site in possible_sites:
                if type(possible_site) == list:
                    site_list += possible_site
                else:
                    site_list.append(possible_site)
            drawing.write_svg(out_file, numbered_atoms=site_list)
        return possible_sites



def draw_line(start_x, end_x, start_y, end_y):
    line = f"""<line x1="{start_x}" x2="{end_x}" y1="{start_y}" y2="{end_y}" stroke-width="2" stroke="black" />"""
    return line


def draw_arrow(start_x, end_x, y, arrow_nr):
    svg = ''
    line_1 = draw_line(start_x, end_x, y, y)
    line_2 = draw_line(end_x - 5, end_x, y - 5, y)
    line_3 = draw_line(end_x - 5, end_x, y + 5, y)

    svg += f"""<g id="arrow_{arrow_nr}">\n"""
    svg += f"{line_1}\n"
    svg += f"{line_2}\n"
    svg += f"{line_3}\n"

    svg += "</g>\n"

    return svg


def make_reaction_text(start_x, end_x, y, text):

    text_x = start_x + (end_x - start_x)/2
    text = f"""<text x="{text_x}" y="{y}" fill="'black" text-anchor="middle" font-family="verdana" font-size = "{12}">\
<tspan y="{y}" dy="0.35em">{text}</tspan></text>"""

    return text


def draw_arrow_and_text(arrow_start, arrow_end, arrow_height, arrow_nr, text):
    arrow_svg = draw_arrow(arrow_start, arrow_end, arrow_height, arrow_nr)
    reaction_text = make_reaction_text(arrow_start, arrow_end, arrow_height - 11, text)

    svg = ''

    svg += f"""<g id="labelled_arrow_{arrow_nr}">\n"""
    svg += arrow_svg
    svg += f"{reaction_text}\n"
    svg += "</g>\n"

    return svg


def draw_double_arrow(arrow_start, arrow_end, arrow_height, arrow_nr):
    arrow_1 = draw_arrow(arrow_start, arrow_end - 10, arrow_height - 5, arrow_nr)
    arrow_2 = draw_arrow(arrow_start + 10, arrow_end, arrow_height + 5, arrow_nr)

    svg = ''

    svg += f"""<g id="labelled_arrow_{arrow_nr}">\n"""
    svg += arrow_1
    svg += arrow_2
    svg += "</g>\n"

    return svg





from pikachu.smiles.smiles import *
from pikachu.general import read_smiles
from raichu.central_chain_detection.label_central_chain import label_pk_central_chain


def reorder_central_chain(central_chain, drawer):
    stop_linearising = None
    rings = []
    full_ring = []
    full_rings = []
    current_ring = []
    current_ring_index = None

    for i, atom in enumerate(central_chain):
        if atom.is_inside_ring(drawer.structure):

            ring_index = atom.get_ring_index(drawer.structure)
            # append last ring once ring is over
            if current_ring and ring_index != current_ring_index:
                rings.append(current_ring)
                full_rings.append(full_ring)
            if current_ring and ring_index == current_ring_index:
                current_ring.append(atom)
                current_ring_index = ring_index
            else:        
                full_ring = atom.get_ring(drawer.structure)
                current_ring = [atom]
                current_ring_index = ring_index

    rings.append(current_ring)
    full_rings.append(full_ring)
    new_rings = []

    for ring in rings:

        if len(ring) > 3:

            alternative_path = drawer.find_shortest_path(drawer.structure.get_atom(ring[0]),
                                                         drawer.structure.get_atom(ring[-1]), path_type='atom')
            if len(alternative_path) <= 3:

                new_backbone = []
                path_added = False
                for atom in central_chain:
                    if atom not in ring:
                        new_backbone.append(atom)
                    elif not path_added:
                        new_backbone += alternative_path
                        path_added = True
                central_chain = new_backbone
                new_rings.append(alternative_path)
            else:
                stop_linearising = ring[1]
                break
        else:
            new_rings.append(ring)

    rings = new_rings
    atom_to_ring = {}
    for ring in rings:
        for atom in ring:
            atom_to_ring[atom] = ring

    return central_chain, full_rings, rings, atom_to_ring, stop_linearising


def find_central_chain(pks_nrps_attached):
    """Identifies the central chain atoms in the input structure from the
    in_central_chain Atom attribute, and returns these Atom objects as a list

    pks_nrps_attached: PIKAChU Structure object, input (hybrid) PK/NRP structure
    """
    # If the structure is a PK starter unit, find the central carbon chain

    if not any(atom.annotations.in_central_chain for atom in pks_nrps_attached.graph) and\
            len(pks_nrps_attached.find_substructures(read_smiles('C(=O)S'))) > 0:
        pks_nrps_attached = label_pk_central_chain(pks_nrps_attached)

    # Find atoms in the structure inside a cycle
    pks_nrps_attached.find_cycles()
    for atom in pks_nrps_attached.graph:
        if atom.in_ring(pks_nrps_attached):
            atom.inside_ring = True
        else:
            atom.inside_ring = False

    # Identify starting point central chain attached NRP/PK
    sulphur = None
    for atom in pks_nrps_attached.graph:
        if atom.type == 'S' and any(neighbour.annotations.domain_type
                                    for neighbour in atom.neighbours):
            sulphur = atom
        if not atom.annotations.in_central_chain:
            atom.annotations.in_central_chain = False
    assert sulphur

    central_chain = [sulphur]
    visited = [sulphur]
    atom_central_chain = sulphur
    end_atom = False

    # Identify complete central chain from in_central_chain Atom attributes
    while not end_atom:

        nothing_changed = True

        for neighbour in atom_central_chain.neighbours:
            if neighbour.annotations.in_central_chain and neighbour not in visited:
                central_chain.append(neighbour)
                visited.append(neighbour)
                atom_central_chain = neighbour
                nothing_changed = False
            elif not neighbour.annotations.in_central_chain:
                neighbours = []
                for next_atom in atom_central_chain.neighbours:
                    neighbours.append(next_atom)
                if all(atom in visited for atom in neighbours):
                    end_atom = True
                else:
                    visited.append(neighbour)
                    nothing_changed = False

        if nothing_changed:
            end_atom = True

    return central_chain


def find_central_chain_ripp(ripp_attached):
    """Identifies the central chain atoms in the input structure from the
    in_central_chain Atom attribute, and returns these Atom objects as a list

    ripp_attached: PIKAChU Structure object, input (hybrid) ripp structure
    """
    # If the structure is a PK starter unit, find the central carbon chain

    if not any(atom.annotations.in_central_chain for atom in ripp_attached.graph) and\
            len(ripp_attached.find_substructures(read_smiles('C(=O)S'))) > 0:
        ripp_attached = label_pk_central_chain(ripp_attached)

    # Find atoms in the structure inside a cycle
    ripp_attached.find_cycles()
    for atom in ripp_attached.graph:
        if atom.in_ring(ripp_attached):
            atom.inside_ring = True
        else:
            atom.inside_ring = False

    # Identify starting point central chain
    nitrogen = None
    domains = [
        atom.annotations.domain_type for atom in ripp_attached.graph if atom.annotations.domain_type]
    if "Follower" in domains:
        for atom in ripp_attached.graph:
            if atom.annotations.domain_type:
                if atom.annotations.domain_type == "Follower":
                    nitrogen = atom.get_neighbour("N")

            if not atom.annotations.in_central_chain:
                atom.annotations.in_central_chain = False

    else:
        for atom in ripp_attached.graph:
            if atom.type == 'N' and any(neighbour.annotations.domain_type
                                        for neighbour in atom.neighbours):
                nitrogen = atom

            if not atom.annotations.in_central_chain:
                atom.annotations.in_central_chain = False

    assert nitrogen

    central_chain = [nitrogen]
    visited = [nitrogen]
    atom_central_chain = nitrogen
    end_atom = False

    # Identify complete central chain from in_central_chain Atom attributes
    while not end_atom:
        for neighbour in atom_central_chain.neighbours:
            if neighbour.annotations.in_central_chain and neighbour not in visited:
                central_chain.append(neighbour)
                visited.append(neighbour)
                atom_central_chain = neighbour
            elif not neighbour.annotations.in_central_chain:
                neighbours = []
                for next_atom in atom_central_chain.neighbours:
                    neighbours.append(next_atom)
                if all(atom in visited for atom in neighbours):
                    end_atom = True
                else:
                    visited.append(neighbour)

    return central_chain


def find_central_chain_not_attached(pks_nrps):
    """Identifies the central chain atoms in the input structure from the
    in_central_chain Atom attribute, and returns these Atom objects as a list

    pks_nrps: PIKAChU Structure object, input (hybrid) PK/NRP/RiPP structure
    """
    # Find atoms in the structure inside a cycle
    pks_nrps.find_cycles()
    for atom in pks_nrps.graph:
        if atom.in_ring(pks_nrps):
            atom.inside_ring = True
        else:
            atom.inside_ring = False

    # Identify starting point central chain attached NRP/PK
    carbon = None
    for atom in pks_nrps.graph:
        if atom.type == 'C' and atom.annotations.in_central_chain and [neighbour.type for neighbour in atom.neighbours].count("O") == 2:
            carbon = atom
        if not atom.annotations.in_central_chain:
            atom.annotations.in_central_chain = False
    assert carbon

    central_chain = [carbon]
    visited = [carbon]
    atom_central_chain = carbon
    end_atom = False

    # Identify complete central chain from in_central_chain Atom attributes
    while not end_atom:
        for neighbour in atom_central_chain.neighbours:
            if neighbour.annotations.in_central_chain and neighbour not in visited:
                central_chain.append(neighbour)
                visited.append(neighbour)
                atom_central_chain = neighbour
            elif not neighbour.annotations.in_central_chain:
                neighbours = []
                for next_atom in atom_central_chain.neighbours:
                    neighbours.append(next_atom)
                if all(atom in visited for atom in neighbours):
                    end_atom = True
                else:
                    visited.append(neighbour)

    return central_chain



from typing import Union, Tuple
from pikachu.chem.structure import Structure
from pikachu.general import read_smiles
from raichu.substrate import NRPSSubstrate, PKSSubstrate
from raichu.reactions.pks_tailoring_reactions import exo_methylen_oxidase, ketoreduction, enoylreduction, dehydration, alpha_L_methyl_transferase, alpha_methyl_transferase, smallest_cyclisation, alpha_hydroxylase, gamma_beta_dehydratase, beta_hydroxy_methyl_transferase, beta_methyl_transferase
from raichu.reactions.nrps_tailoring_reactions import epimerize, n_methylate, nrps_cyclodehydration, nrps_oxidation
from raichu.reactions.pks_elongation_reactions import pks_elongation
from raichu.reactions.nrps_elongation_reactions import nrps_elongation
from raichu.reactions.chain_release import release_linear_reduction, release_linear_thioesterase
from raichu.domain.domain_types import DomainSuperClass, RecognitionDomainType, CarrierDomainType, \
    TailoringDomainType, SynthesisDomainType, TerminationDomainType, KSDomainSubtype, KRDomainSubtype, ERDomainSubtype

from dataclasses import dataclass


@dataclass
class Domain:
    supertype: DomainSuperClass
    type: Union[TailoringDomainType, CarrierDomainType, SynthesisDomainType, RecognitionDomainType,
                TerminationDomainType]
    subtype: Union[None, KRDomainSubtype, KSDomainSubtype, ERDomainSubtype]
    domain_name: Union[str, None]
    active: bool = True
    gene: Union[str, None] = None
    used: bool = True

    def __post_init__(self):
        if self.domain_name is None:
            self.domain_name = self.type.name

    def __repr__(self):
        return self.domain_name

    def set_gene(self, gene_name: str) -> None:
        self.gene = gene_name


class UnknownDomain(Domain):
    def __init__(self, domain_name: str, active: bool = True) -> None:
        superclass = DomainSuperClass.from_string("TAILORING")
        domain_type = TailoringDomainType.from_string("UNKNOWN")
        assert domain_name
        super().__init__(superclass, domain_type, None, domain_name, active=active, used=False)


class TailoringDomain(Domain):
    def __init__(self, domain_type: str, domain_subtype: Union[str, None] = None, active: bool = True,
                 domain_name: Union[str, None] = None, used: bool = True) -> None:
        superclass = DomainSuperClass.from_string("TAILORING")
        domain_type = TailoringDomainType.from_string(domain_type)
        if domain_subtype is not None:
            if domain_type.name == 'KR' or domain_type.name == 'DUMMY_KR':
                domain_subtype = KRDomainSubtype.from_string(domain_subtype)
            elif domain_type.name == 'ER' or domain_type.name == 'DUMMY_ER':
                domain_subtype = ERDomainSubtype.from_string(domain_subtype)
            else:
                raise ValueError(f"RAIChU does not support domain subtypes for {domain_type.name}")
        else:
            if domain_type.name == 'KR' or domain_type.name == 'DUMMY_KR':
                domain_subtype = KRDomainSubtype.from_string("UNKNOWN")
            elif domain_type.name == 'ER' or domain_type.name == 'DUMMY_ER':
                domain_subtype = ERDomainSubtype.from_string("UNKNOWN")

        super().__init__(superclass, domain_type, domain_subtype, domain_name, active=active, used=used)

    def do_tailoring(self, structure) -> Tuple[Structure, bool]:
        """
        Performs tailoring reaction
        """
        if self.type.name == 'KR' or self.type.name == 'DUMMY_KR':
            return ketoreduction(structure, self.subtype)
        elif self.type.name == 'DH' or self.type.name == 'DUMMY_DH':
            return dehydration(structure)
        elif self.type.name == 'EDH' or self.type.name == 'DUMMY_EDH':
            return dehydration(structure, "E")
        elif self.type.name == 'ZDH' or self.type.name == 'DUMMY_ZDH':
            return dehydration(structure, "Z")
        elif self.type.name == 'ER' or self.type.name == 'DUMMY_ER':
            return enoylreduction(structure, self.subtype)
        elif self.type.name == 'ALMT' or self.type.name == 'DUMMY_ALMT':
            return alpha_L_methyl_transferase(structure)
        elif self.type.name == 'AMT' or self.type.name == 'DUMMY_AMT':
            return alpha_methyl_transferase(structure)
        elif self.type.name == 'SC' or self.type.name == 'DUMMY_SC':
            return smallest_cyclisation(structure)
        elif self.type.name == 'AH' or self.type.name == 'DUMMY_AH':
            return alpha_hydroxylase(structure)
        elif self.type.name == 'GDH' or self.type.name == 'DUMMY_GDH':
            return gamma_beta_dehydratase(structure)
        elif self.type.name == 'EGDH' or self.type.name == 'DUMMY_EGDH':
            return gamma_beta_dehydratase(structure, "E")
        elif self.type.name == 'ZGDH' or self.type.name == 'DUMMY_ZGDH':
            return gamma_beta_dehydratase(structure, "Z")
        elif self.type.name == 'OMT' or self.type.name == 'DUMMY_OMT':
            return beta_hydroxy_methyl_transferase(structure)
        elif self.type.name == 'BMT' or self.type.name == 'DUMMY_BMT':
            return beta_methyl_transferase(structure)
        elif self.type.name == 'EMO' or self.type.name == 'DUMMY_EMO':
            return exo_methylen_oxidase(structure)
        elif self.type.name == 'E':
            return epimerize(structure)
        elif self.type.name == 'nMT':
            return n_methylate(structure)
        elif self.type.name == 'CYC':
            return nrps_cyclodehydration(structure)
        elif self.type.name == 'OX':
            return nrps_oxidation(structure)
        else:
            raise Warning(f"Tailoring domain {self.domain_name} not recognised by RAIChU. Ignored.")


class SynthesisDomain(Domain):
    def __init__(self, domain_type: str, domain_subtype: Union[str, None] = None, active: bool = True,
                 domain_name: Union[str, None] = None, is_elongating: bool = True,
                 used: bool = True) -> None:
        superclass = DomainSuperClass.from_string("SYNTHESIS")
        domain_type = SynthesisDomainType.from_string(domain_type)

        if domain_subtype is not None:
            if domain_type.name == 'KS' or domain_type.name == "DUMMY_KS":
                domain_subtype = KSDomainSubtype.from_string(domain_subtype)
            else:
                raise ValueError(f"RAIChU does not support domain subtypes for {domain_type.name}")

        super().__init__(superclass, domain_type, domain_subtype, domain_name, active=active, used=used)
        self.is_elongating = is_elongating

    def do_elongation(self, structure, substrate) -> Structure:
        """
        Performs elongation reaction
        """
        if self.type.name == 'C' or self.type.name == "DUMMY_C":
            if self.is_elongating:
                building_block = read_smiles(substrate.smiles)
                return nrps_elongation(building_block, structure)
            else:
                return structure
        elif self.type.name == 'KS' or self.type.name == "DUMMY_KS":
            if self.is_elongating:
                if self.subtype is None or self.subtype.name in [v.name for v in KSDomainSubtype]:
                    building_block = substrate.elongation_monomer
                    return pks_elongation(structure, building_block)
                else:
                    raise ValueError(f"RAIChU does not support domain subtype {self.subtype.name}")
            else:
                return structure
        else:
            raise ValueError(f"Unknown type of synthesis domain: {self.type}")


class RecognitionDomain(Domain):
    def __init__(self, domain_type: str, substrate_name: str, domain_subtype: Union[str, None] = None,
                 active: bool = True, domain_name: Union[str, None] = None, used: bool = True) -> None:

        superclass = DomainSuperClass.from_string("RECOGNITION")
        domain_type = RecognitionDomainType.from_string(domain_type)

        if domain_subtype is not None:
            raise ValueError(f"RAIChU does not support domain subtypes for {domain_type.name}")

        super().__init__(superclass, domain_type, domain_subtype, domain_name, active=active, used=used)

        if self.type.name == 'A' or self.type.name == 'DUMMY_A' or self.type.name == 'CAL':
            self.substrate = NRPSSubstrate(substrate_name)
        elif self.type.name == 'AT' or self.type.name == 'DUMMY_AT':
            self.substrate = PKSSubstrate(substrate_name)
        else:
            raise ValueError(f"Unknown type of recognition domain: {self.type}")

    @property
    def smiles(self):
        return self.substrate.smiles

    @property
    def substrate_name(self):
        return self.substrate.name


class CarrierDomain(Domain):
    def __init__(self, domain_type: str, domain_subtype: Union[str, None] = None,
                 domain_name: Union[str, None] = None, active: bool = True,
                 used: bool = True) -> None:
        superclass = DomainSuperClass.from_string("CARRIER")
        domain_type = CarrierDomainType.from_string(domain_type)

        if domain_subtype is not None:
            raise ValueError(f"RAIChU does not support domain subtypes for {domain_type.name}")

        super().__init__(superclass, domain_type, domain_subtype, domain_name, active=active, used=used)


class TerminationDomain(Domain):
    def __init__(self, domain_type: str, domain_subtype: Union[str, None] = None, active: bool = True,
                 domain_name: Union[str, None] = None, used: bool = True) -> None:
        superclass = DomainSuperClass.from_string("TERMINATION")
        domain_type = TerminationDomainType.from_string(domain_type)

        if domain_subtype is not None:
            raise ValueError(f"RAIChU does not support domain subtypes for {domain_type.name}")

        super().__init__(superclass, domain_type, domain_subtype, domain_name, active=active, used=used)

    def release_chain(self, structure) -> Structure:
        """
        Performs chain release reaction and returns True if reaction was successful, False otherwise
        """
        if self.type.name == "TE" or self.type.name == "DUMMY_TE":
            return release_linear_thioesterase(structure)
        elif self.type.name == "TD" or self.type.name == "DUMMY_TD":
            return release_linear_reduction(structure)
        else:
            raise ValueError(f"Unknown type of termination domain: {self.type}")



from pikachu.reactions.functional_groups import GroupDefiner, find_atoms, find_bonds
from raichu.reactions.general import initialise_atom_attributes

import timeout_decorator
from raichu.data.molecular_moieties import AMINO_FATTY_ACID, AMINO_ACID_BACKBONE, N_AMINO_ACID, C1_AMINO_ACID, \
    C2_AMINO_ACID, BETA_AMINO_ACID_BACKBONE, B_N_AMINO_ACID, B_C1_AMINO_ACID, B_C2_AMINO_ACID, B_C3_AMINO_ACID, \
    B_LEAVING_BOND, ACID_C1, OH_AMINO_ACID, N_AMINO_ACID_ATTACHED, C1_AMINO_ACID_ATTACHED, C2_AMINO_ACID_ATTACHED

POLYKETIDE_S = GroupDefiner('Sulphur atom polyketide', 'SC(C)=O', 0)
POLYKETIDE_S_INSERTED_O = GroupDefiner('Sulphur atom polyketide inserted O', 'SC(O)=O', 0)
POLYKETIDE_S_REDUCED_STARTER = GroupDefiner('Sulphur atom polyketide', 'SC(C)O', 0)


class CentralChain:
    def __init__(self, structure, chain_type):
        self.structure = structure
        if chain_type in ['ripp', 'nrps', 'pks']:
            self.chain_type = chain_type
        else:
            raise ValueError(f"Expected 'ripp', 'nrps', or 'pks' as central chain type; got {chain_type}")

    def label_central_chain(self):
        raise NotImplementedError


class PksCentralChain(CentralChain):
    def __init__(self, structure):
        super().__init__(structure, chain_type='pks')


class NrpsCentralChain(CentralChain):
    def __init__(self, structure):
        super().__init__(structure, chain_type='nrps')


class RippCentralChain(CentralChain):
    def __init__(self, structure):
        super().__init__(structure, chain_type='ripp')


@timeout_decorator.timeout(3)
def label_pk_central_chain(pk_starter_unit):
    """Finds the the atoms in the central chain of the polyketide starter unit,
    sets the in_central_chain Atom object attribute to True/False accordingly,
    after which the pk_starter_unit Structure object is returned.

    pk_starter_unit: PIKAChU Structure object of the PK starter unit
    """

    initialise_atom_attributes(pk_starter_unit)
    central_chain = []
    visited = []

    # Check if structure is a polyketide, define sulphur attached to domain
    locations_sulphur = find_atoms(POLYKETIDE_S, pk_starter_unit)
    if len(locations_sulphur) == 0:
        locations_sulphur = find_atoms(POLYKETIDE_S_INSERTED_O, pk_starter_unit)
        if len(locations_sulphur) == 0:
            locations_sulphur = find_atoms(POLYKETIDE_S_REDUCED_STARTER, pk_starter_unit)
    assert len(locations_sulphur) == 1
    
    sulphur_polyketide = locations_sulphur[0]
    central_chain.append(sulphur_polyketide)

    # Iterate over atom neighbours, check if the atom belongs to the central
    # carbon chain (i.e., not a methyl/ethyl/methoxy sidechain) and add to list
    for neighbour in sulphur_polyketide.neighbours:
        if neighbour.type == 'C':
            chain_carbon = neighbour
            end_carbon = False
            central_chain.append(chain_carbon)
            visited.append(chain_carbon)

            while not end_carbon:

                for next_atom in chain_carbon.neighbours:
                    ethyl_branch = False
                    inside_cycle = False
                    methyl_group = False

                    # Build lists of neighbouring atom types
                    if next_atom.type == 'C' and next_atom not in visited or \
                            (next_atom.type == 'O' and next_atom not in visited and
                             len(next_atom.get_neighbours('C')) == 2):

                        # Keep track of carbon neighbours

                        # Confirm carbon doesn't belong to ethyl sidebranch
                        types = []
                        c_neighbours = next_atom.get_neighbours('C')
                        if len(c_neighbours) == 2 and next_atom.get_neighbour('O') is None:

                            for carbon_neighbour in c_neighbours:
                                for branch_neighbour in carbon_neighbour.neighbours:
                                    types.append(branch_neighbour.type)

                            if types.count('H') == 4 and types.count('C') == 4:
                                ethyl_branch = True

                            for carbon_neighbour in c_neighbours:
                                if carbon_neighbour not in visited and carbon_neighbour != next_atom:
                                    ethyl_branch = False

                            visited.append(chain_carbon)

                        # Carbon in a terminal carboxylic acid group is the
                        # final carbon in the central chain
                        if len(next_atom.get_neighbours('O')) == 2:
                            for oxygen in next_atom.get_neighbours('O'):
                                if next_atom.get_bond(oxygen).type == 'single':
                                    central_chain.append(oxygen)
                                    visited.append(oxygen)
                            central_chain.append(next_atom)
                            end_carbon = True
                            visited.append(chain_carbon)

                        # Confirm carbon is not part of cycle
                        if len(c_neighbours) == 2:
                            if all(carbon.in_ring(pk_starter_unit) for carbon in c_neighbours):
                                visited.append(next_atom)
                                inside_cycle = True
                                end_carbon = True

                        # Case if carbon is part of a benzene ring
                        if len(c_neighbours) == 3 and len(next_atom.neighbours) == 3:
                            # central_chain.append(next_atom) #added 10/04/2021
                            visited.append(next_atom)
                            inside_cycle = True
                            end_carbon = True

                        # Case where the PK starter ends in two methyl branches
                        if len(next_atom.get_neighbours('C')) == 3 and len(next_atom.get_neighbours('H')) == 1:
                            nr_methyl_brances = 0
                            methyl_carbon = None
                            for c_neighbour in c_neighbours:
                                c_neighbour_neighbour_types = []
                                for c_neighbour_neighbour \
                                        in c_neighbour.neighbours:
                                    c_neighbour_neighbour_types.append(
                                        c_neighbour_neighbour.type)
                                if c_neighbour_neighbour_types.count('H') == 3:
                                    methyl_carbon = c_neighbour
                                    nr_methyl_brances += 1

                            if nr_methyl_brances == 2:
                                assert methyl_carbon is not None
                                methyl_group = True
                                visited.append(next_atom)
                                central_chain.append(next_atom)
                                visited.append(methyl_carbon)
                                central_chain.append(methyl_carbon)
                                end_carbon = True

                        # Confirm carbon doesn't belong to methyl sidebranch
                        if len(next_atom.get_neighbours('H')) == 3 or \
                                (len(next_atom.get_neighbours('H')) == 2 and len(next_atom.get_neighbours('*')) == 1):
                            methyl_group = True

                            # Terminal methyl group is not a side branch:
                            count_c_neighbours_not_visited = 0
                            for neighbouring_atom in chain_carbon.neighbours:
                                if neighbouring_atom.type == 'C' and neighbouring_atom not in visited:
                                    count_c_neighbours_not_visited += 1

                            if count_c_neighbours_not_visited < 2:
                                methyl_group = False
                                visited.append(chain_carbon)
                                for carbon_neighbour in chain_carbon.neighbours:
                                    if carbon_neighbour.type == 'C' and carbon_neighbour not in visited:
                                        next_neighbours = []
                                        for next_neighbour in carbon_neighbour.neighbours:
                                            next_neighbours.append(
                                                next_neighbour.type)
                                        if next_neighbours.count('H') == 3 or (
                                                next_neighbours.count(
                                                        'H') == 2 and (
                                                next_neighbours.count(
                                                        '*')) == 1):
                                            central_chain.append(carbon_neighbour)
                                end_carbon = True

                            visited.append(chain_carbon)

                        # Identification of central chain carbon atoms through
                        # exclusion
                        if not ethyl_branch and not methyl_group and not \
                                end_carbon and not inside_cycle:
                            central_chain += [next_atom]
                            chain_carbon = next_atom
                            visited.append(chain_carbon)

    # Set in_central_chain Atom attribute
    for atom in pk_starter_unit.graph:
        if atom in central_chain:
            atom.annotations.in_central_chain = True
        else:
            atom.annotations.in_central_chain = False


def label_ripp_central_chain(peptide):
    initialise_atom_attributes(peptide)

    n_atoms_aa = find_atoms(N_AMINO_ACID_ATTACHED, peptide)
    c1_atoms_aa = find_atoms(C1_AMINO_ACID_ATTACHED, peptide)
    c2_atoms_aa = find_atoms(C2_AMINO_ACID_ATTACHED, peptide)
    oh_atoms_aa = find_atoms(OH_AMINO_ACID, peptide)

    for nitrogen in n_atoms_aa:
        nitrogen.annotations.in_central_chain = True
        nitrogen.annotations.n_atom_nmeth = True
    for carbon in c1_atoms_aa:
        carbon.annotations.in_central_chain = True
        carbon.annotations.chiral_c_ep = True
    for carbon in c2_atoms_aa:
        carbon.annotations.in_central_chain = True
    for oxygen in oh_atoms_aa:
        oxygen.annotations.in_central_chain = True


def label_nrp_central_chain(peptide, module_type='elongation'):
    initialise_atom_attributes(peptide)

    as_normal = False
    if module_type == 'starter':
        if peptide.find_substructures(AMINO_FATTY_ACID):
            acid = peptide
            label_acid_central_chain(acid)
        else:
            as_normal = True

    if as_normal or module_type == 'elongation':
        if peptide.find_substructures(AMINO_ACID_BACKBONE):
            n_atoms_aa = find_atoms(N_AMINO_ACID, peptide)
            c1_atoms_aa = find_atoms(C1_AMINO_ACID, peptide)
            c2_atoms_aa = find_atoms(C2_AMINO_ACID, peptide)

            assert len(n_atoms_aa) == 1
            assert len(c1_atoms_aa) == 1
            assert len(c2_atoms_aa) == 1

            n_atoms_aa[0].annotations.in_central_chain = True
            n_atoms_aa[0].annotations.n_atom_nmeth = True
            c1_atoms_aa[0].annotations.in_central_chain = True
            c1_atoms_aa[0].annotations.chiral_c_ep = True
            c2_atoms_aa[0].annotations.in_central_chain = True

        elif peptide.find_substructures(BETA_AMINO_ACID_BACKBONE):
            n_atoms_aa = find_atoms(B_N_AMINO_ACID, peptide)
            c1_atoms_aa = find_atoms(B_C1_AMINO_ACID, peptide)
            c2_atoms_aa = find_atoms(B_C2_AMINO_ACID, peptide)
            c3_atoms_aa = find_atoms(B_C3_AMINO_ACID, peptide)

            assert len(n_atoms_aa) == 1
            assert len(c1_atoms_aa) == 1
            assert len(c2_atoms_aa) == 1
            assert len(c3_atoms_aa) == 1

            if n_atoms_aa[0].in_ring(peptide) and c1_atoms_aa[0].in_ring(peptide) and c2_atoms_aa[0].in_ring(peptide):
                label_acid_central_chain(peptide)

            else:

                n_atoms_aa[0].annotations.in_central_chain = True
                n_atoms_aa[0].annotations.n_atom_nmeth = True
                c1_atoms_aa[0].annotations.in_central_chain = True
                c1_atoms_aa[0].annotations.chiral_c_ep = True
                c2_atoms_aa[0].annotations.in_central_chain = True
                c3_atoms_aa[0].annotations.in_central_chain = True

                bonds = find_bonds(B_LEAVING_BOND, peptide)
                assert len(bonds) == 1
                bond = bonds[0]
                if bond.atom_1.type == 'O':
                    bond.atom_1.annotations.leaving_oh_o = True

                    bond.atom_2.annotations.leaving_oh_h = True

                elif bond.atom_2.type == 'O':
                    bond.atom_2.annotations.leaving_oh_o = True

                    bond.atom_1.annotations.leaving_oh_h = True

        else:
            # If NRPS starter unit is a (fatty) acid instead of an amino acid
            acid = peptide
            label_acid_central_chain(acid)


def label_acid_central_chain(acid):
    c1_atom = None
    c2_atom = None
    oh_o_atom = None
    oh_h_atom = None
    backbone = []

    c1_atoms_aa = find_atoms(ACID_C1, acid)
    if c1_atoms_aa:
        c1_atom = c1_atoms_aa[0]

    assert c1_atom

    for neighbour in c1_atom.neighbours:
        neighbours = []

        if neighbour.type == 'C':
            for next_atom in neighbour.neighbours:
                neighbours.append(next_atom.type)
            if neighbours.count('O') == 2 and len(neighbours) == 3:
                c2_atom = neighbour

                # Find the leaving -OH group and annotate it

                for bond in c2_atom.bonds:
                    if bond.type == 'single' and (bond.has_neighbour('O')):
                        if bond.atom_1.type == 'O':
                            bond.atom_1.annotations.leaving_oh_o = True
                            oh_o_atom = bond.atom_1

                            bond.atom_2.annotations.leaving_oh_h = True
                            oh_h_atom = bond.atom_2

                        elif bond.atom_2.type == 'O':
                            bond.atom_2.annotations.leaving_oh_o = True
                            oh_o_atom = bond.atom_2

                            bond.atom_1.annotations.leaving_oh_h = True
                            oh_h_atom = bond.atom_1

    assert c2_atom
    assert oh_o_atom
    assert oh_h_atom

    c1_atom.annotations.in_central_chain = True
    c2_atom.annotations.in_central_chain = True
    c2_atom.annotations.c2_acid = True

    backbone.append(c1_atom)
    backbone.append(c2_atom)

    last_backbone_atom = c1_atom
    endpoint = False

    # Add atoms in chain with exactly 2 non-H neighbours to the central chain
    while not endpoint:

        neighbours = last_backbone_atom.get_non_hydrogen_neighbours()[:]
        for neighbour in last_backbone_atom.get_non_hydrogen_neighbours():
            if neighbour in backbone:
                neighbours.remove(neighbour)

        if not neighbours:
            endpoint = True

        else:
            # If the last backbone atom was the first ring member
            if last_backbone_atom.in_ring(acid):
                ring_members = last_backbone_atom.get_ring(acid)
                outgoing_pairs = []
                for atom in ring_members:
                    for neighbour in atom.get_non_hydrogen_neighbours():
                        if neighbour not in ring_members and neighbour not in backbone and not neighbour.in_ring(acid):
                            outgoing_pairs.append((atom, neighbour))

                if not outgoing_pairs:
                    endpoint = True

                else:
                    longest_index = -1
                    largest_subtree = 0
                    best_shortest_path = None
                    for i, pair in enumerate(outgoing_pairs):
                        ring_atom, outgoing_atom = pair
                        shortest_path = find_shortest_path_ring(ring_atom, last_backbone_atom, ring_members)
                        subtree_size = acid.get_subtree_size(outgoing_atom, set(ring_members))
                        if len(shortest_path) < 4 and subtree_size > largest_subtree:
                            longest_index = i
                            best_shortest_path = shortest_path

                    if longest_index >= 0:

                        backbone_ring_atom, next_backbone_atom = outgoing_pairs[longest_index]
                        for atom in best_shortest_path:
                            shortest_path_atom = acid.get_atom(atom)
                            if shortest_path_atom not in backbone:
                                shortest_path_atom.annotations.in_central_chain = True
                                backbone.append(shortest_path_atom)
                            next_backbone_atom = acid.get_atom(next_backbone_atom)
                            next_backbone_atom.annotations.in_central_chain = True
                            backbone.append(next_backbone_atom)
                            last_backbone_atom = next_backbone_atom
                    else:
                        endpoint = True

            else:
                longest_index = -1
                largest_subtree = 0

                for i, neighbour in enumerate(neighbours):
                    subtree_size = acid.get_subtree_size(neighbour, set(backbone))
                    if subtree_size > largest_subtree:
                        largest_subtree = subtree_size
                        longest_index = i

                if longest_index >= 0:
                    next_backbone_atom = acid.get_atom(neighbours[longest_index])
                else:
                    next_backbone_atom = acid.get_atom(neighbours[0])

                next_backbone_atom.annotations.in_central_chain = True
                backbone.append(next_backbone_atom)
                last_backbone_atom = next_backbone_atom


def find_shortest_path_ring(atom_1, atom_2, atom_list):
    max_distance = len(atom_list) // 2
    atom_1_idx = None
    atom_2_idx = None
    for i, atom in enumerate(atom_list):
        if atom == atom_1:
            atom_1_idx = i
        if atom == atom_2:
            atom_2_idx = i

    assert atom_1_idx is not None
    assert atom_2_idx is not None

    large_index = max([atom_1_idx, atom_2_idx])
    small_index = min([atom_1_idx, atom_2_idx])

    shortest_path = atom_list[small_index:large_index + 1]
    if len(shortest_path) - 1 > max_distance:
        shortest_path = atom_list[large_index:] + atom_list[:small_index + 1]

    return shortest_path



from pikachu.reactions.functional_groups import find_atoms
from pikachu.chem.chirality import same_chirality
from raichu.data.molecular_moieties import METHYLMALONYL_SIDECHAIN_CARBON, \
    METHYLMALONYL_SULPHUR_CARBON, METHYLMALONYL_MAINCHAIN_CARBON, METHYLMALONYL_CHIRAL_CARBON


def set_sidechain_chirality(structure):
    chiral_carbons = find_atoms(METHYLMALONYL_CHIRAL_CARBON, structure)
    if len(chiral_carbons) == 1:
        chiral_carbon = chiral_carbons[0]
        sulphur_carbon = find_atoms(METHYLMALONYL_SULPHUR_CARBON, structure)[0]
        sidechain_carbon = find_atoms(METHYLMALONYL_SIDECHAIN_CARBON, structure)[0]
        mainchain_carbon = find_atoms(METHYLMALONYL_MAINCHAIN_CARBON, structure)[0]
        if chiral_carbon.has_neighbour('H'):
            hydrogen = chiral_carbon.get_neighbour('H')
            counterclockwise_order = [hydrogen, sidechain_carbon, sulphur_carbon, mainchain_carbon]
            if same_chirality(counterclockwise_order, chiral_carbon.neighbours):
                chiral_carbon.chiral = 'counterclockwise'
            else:
                chiral_carbon.chiral = 'clockwise'